/*
 * renum.cpp 
 * Renumbers and lints an NFO file specified on the commandline or piped
 * in via standard input.
 *
 * Return values:
 * See README[.txt]
 *
 * Copyright 2004-2006 by Dale McCoy.
 * dalestan@gmail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include<iostream>
#include<string>
#include<sstream>
#include<iomanip>
#include<fstream>
#include<cerrno>
#include<cassert>

using namespace std;

#include"renum.h"
#include"globals.h"
#include"inlines.h"
#include"sanity.h"
#include"command.h"
#include"messages.h"
#include"inject.h"
#ifdef _WIN32
#   include "win32.h"
#endif

#ifndef _MSC_VER
#undef __cdecl
#define __cdecl
#endif

int process_file(istream&);
void output_buffer(const string&,int&,bool);
void comment_sprite(const string&);
bool verify_real(string&);

static int _retval=EOK;

void SetCode(int x){_retval=max(_retval,x);}

void doexit(){exit(_retval);}

void ShowHelp(){
    cout<<"Usage: renum [options] [file1 [file2 [file3 [...]]]]\n"
        "   Any number of files may be specified.\n"
        "   If no files or options are specified, NFORenum will read a file from\n"
        "   standard input and write it to standard output.\n"
        "Options:\n"
        "   --data[=<dir>] -D[<dir>]\n"
        "       If <dir> is specified, look for the .renum directory in <dir>, and\n"
        "       create it if not found. Regardless of whether or not <dir> is\n"
        "       specified, eliminate the 5-second wait used to ensure that those not\n"
        "       running NFORenum from a command line can see the directory-created\n"
        "       message.\n"
        "       Default: Look for  the .renum directory in the current directory, and\n"
        "       then in the environment variable HOME, if defined. If not found attempt\n"
        "       to create in HOME, then in .\n"
        "   --no-replace --keep-old -k\n"
        "       Do not replace the old NFO file; write new file to file.new.nfo.\n"
        "       Default: Use file[.nfo].new as temporary, rename it to file[.nfo]\n"
        "       when done.\n"
        "   --comments=<type> -c <type>\n"
        "       Type is one character, either /, ;, or #, and specifies the comment\n"
        "       style that NFORenum will use. This will not change the header, because\n"
        "       grfcodec requires that the header be commented in C++-style.\n"
        "The following options cause NFORenum to behave as if all files started with\n"
        "the associated command. The readme has full details on the comment commands.\n"
        "Options associated with comment commands that require additional information\n"
        "take that information as an argument. With the exception of -L/--let, the\n"
        "options to the command line versions are case insensitive.\n"
        "   --diff -d                   @@DIFF\n"
        "   --let -L                    @@LET\n"
        "   --lint -l                   @@LINT\n"
        "   --preserve-messages -p      @@PRESERVEMESSAGES\n"
        "   --real-sprites -r           @@REAL\n"
        "   --warning-disable -w        @@WARNING DISABLE\n"
        "   --warning-enable -W         @@WARNING ENABLE\n"
        "       -w and -W (and their long counterparts) also accept a comma-separated\n"
        "       list of messages, all of which will be ENABLEd or DISABLEd.\n\n"
        "NFORenum is Copyright 2004-2006 by Dale McCoy (dalestan@gmail.com)\n"
        "You may copy and redistribute it under the terms of the GNU General Public\n"
        "License, as stated in the file 'COPYING'.\n";
}

int __cdecl main(int argc,char**argv){
    string infilename,outfilename,bakfilename,basename;
    int result,longind,opt=0,replace=1;
    static const option optlist[]={
        {"data",optional_argument,NULL,'D'},
        {"no-replace",no_argument,&replace,0},
        {"keep-old",no_argument,&replace,0},
        {"comments",required_argument,NULL,'c'},
        {"help",no_argument,NULL,'?'},
        {"diff",no_argument,NULL,'d'},
        {"let",required_argument,NULL,'L'},
        {"lint",optional_argument,NULL,'l'},
        {"preserve-messages",no_argument,NULL,'P'},
        {"warning-disable",required_argument,NULL,'w'},
        {"warning-enable",required_argument,NULL,'W'},
        {NULL,0,0,0}
    };
    pOut=argc==1?&cerr:&cout;
    IssueMessage(0,STARTUP);
    ifstream fin;
    ofstream fout;
    while(argc>1){
        if(opt!=EOF)opt=getopt_long(argc,argv,"D::k?c:" "dL:l:pw:W:r:",optlist,&longind);
        switch(opt){
        case 0:continue;
        case'D':
            if(optarg)datadir=optarg;
            dosleep=false;
            continue;
        case'k':replace=0;continue;
        case'c':
            switch(*optarg){
            case'#':COMMENT_PREFIX="#";break;
            case';':COMMENT_PREFIX=";";break;
            case'/':COMMENT_PREFIX="//";
            }
            continue;
        case'?':
            ShowHelp();
            return 0;
        case EOF:
            if(optind==argc)doexit();
            basename=argv[optind++];break;
        default:CLCommand(opt);continue;
        }
        pNfo=&fout;
        bakfilename=basename+bak_ext;
        fin.open((infilename=basename).c_str());
        if(!fin.is_open()){
            fin.open((infilename=basename+nfo_ext).c_str());
            bakfilename=basename+nfo_ext+bak_ext;
        }
        if(fin.is_open()){
            fout.open((outfilename=(replace?basename+new_ext:basename+new_ext+nfo_ext)).c_str());
        }else{
            bakfilename=dirname+(basename+bak_ext);
            fin.open((infilename=dirname+basename).c_str());
            if(!fin.is_open()){
                bakfilename=dirname+(basename+nfo_ext+bak_ext);
                fin.open((infilename=dirname+basename+nfo_ext).c_str());
            }
            if(fin.is_open()){
                fout.open((outfilename=dirname+(replace?basename+new_ext:basename+new_ext+nfo_ext)).c_str());
            }else{
                IssueMessage(0,NO_INPUT_FILE,basename.c_str());
                SetCode(EFILE);
                continue;
            }
        }
        if(!fout.is_open()){
            IssueMessage(0,NO_OUTPUT_FILE,outfilename.c_str(),basename.c_str());
            SetCode(EFILE);
            fin.close();
            fout.clear();
            continue;
        }
        fin.clear();
        reset_sanity();
        reset_commands();
        IssueMessage(0,PROCESSING_FILE,basename.c_str());
        result=process_file(fin);
        fin.close();
        fout.close();
        if(result)unlink(outfilename.c_str());
        else if(replace){
            if(rename(infilename.c_str(),bakfilename.c_str())&&errno==EEXIST)unlink(infilename.c_str());
            if(rename(outfilename.c_str(),infilename.c_str())){
                IssueMessage(0,REPLACE_FAILED,infilename.c_str(),outfilename.c_str(),errno);
                SetCode(EFILE);
            }
        }
        IssueMessage(0,PROCESSING_COMPLETE);
    }
    pNfo=&cout;
    IssueMessage(0,PROCESSING);
    process_file(cin);
    IssueMessage(0,PROCESSING_COMPLETE);
    return _retval;
}

#define flush_buffer()\
    do{\
        if(buffer!=""){\
            output_buffer(buffer,spritenum,isPatch);\
            buffer="";\
        }\
    }while(0)

#define SetVersion(x)\
    (NFOversion=max((x),NFOversion))

static int NFOversion;

enum{INVALID_CHAR=1,UNTERM_STR,OFFSET_STR};

string smash(const string&,int&);

int process_file(istream&in){
    NFOversion=4;
    string sprite,datapart,buffer;
    inject_into(in);

//This section is a rewrite of lines 35-50 or thereabouts of info.cc
//from grfcodec: http://www.ttdpatch.net/grfcodec
//grfcodec is Copyright 2000-2003 Josef Drexler
    getline(in,sprite);//Info header or first sprite
    if (sprite.substr(0,3)=="// "){
        getline(in,sprite);//Info version (first sprite for ancient NFO versions)
        if(sscanf(sprite.c_str(),"// (Info version %d)",&NFOversion)){
            if(NFOversion<4||NFOversion>6){
                IssueMessage(0,UNKNOWN_VERSION,NFOversion);
                if(NFOversion>6){
                    IssueMessage(0,SKIPPING_FILE);
                    return-1;
                }
                IssueMessage(0,PARSING_FILE);
            }
            if(NFOversion>2)
                getline(in,sprite);//Format line
        }else{
            IssueMessage(0,UNKNOWN_VERSION,1);
            IssueMessage(0,PARSING_FILE);
        }
        getline(in,sprite);//first sprite
    }


    int temp,spritenum=-1;
    string::size_type firstnotpseudo;
    bool isPatch=false;
    ostringstream outbuffer;
    ostream*real_out=pNfo;
    pNfo=&outbuffer;
    SetVersion(4);
    AutoConsoleMessages(spritenum);
    while(true){
        //IssueMessage(0,SPRITE,spritenum+1,sprite.c_str());
        istringstream spritestream(sprite);
        eat_white(spritestream);
        if(spritestream.peek()==EOF)buffer+='\n';
        else if(COMMENT.find((char)spritestream.peek())!=NPOS){//comment
            if(is_command(sprite))
                flush_buffer();
            if(parse_comment(sprite))
                buffer+=sprite+'\n';
        }else{//sprite
            if(!eat_white(spritestream>>temp))spritestream.clear();
            if(spritestream.peek()=='*'){
                if(spritestream.ignore().peek()=='*'){
                    SetVersion(6);
                    getline(eat_white(spritestream.ignore()),datapart);
                    flush_buffer();
                    if(isPatch)check_sprite(INCLUDE);
                    else IssueMessage(ERROR,UNEXPECTED,BIN_INCLUDE);
                    spritenum++;
                    (*pNfo)<<setw(5)<<(GetState(DIFF)?-1:spritenum)<<" **\t "<<datapart<<'\n';
                }else{
                    eat_white(spritestream>>temp);
                    flush_buffer();
                    if(spritenum==-1){
                        isPatch=true;
                        if(temp!=4)
                            outbuffer<<COMMENT_PREFIX<<sprite<<endl;
                        spritenum=0;
                    }else{
                        getline(spritestream,buffer);
                        buffer+='\n';
                    }
                }
            }else{
                getline(spritestream,datapart);
                firstnotpseudo=datapart.find_first_not_of(VALID_PSEUDO);
                if(!spritestream||firstnotpseudo==NPOS||(datapart[firstnotpseudo]=='"'?(SetVersion(5),true):false)||
                    COMMENT.find_first_of(datapart[firstnotpseudo])!=NPOS)
                    buffer+=sprite+'\n';
                else{
                    spritenum++;
                    if(verify_real(datapart)){
                        spritenum--;
                        flush_buffer();
                        spritenum++;
                        if(isPatch)check_sprite(REAL);
                        (*pNfo)<<setw(5)<<(GetState(DIFF)?-1:spritenum)<<' '<<datapart<<endl;
                    }else{
                        buffer+="//"+sprite+'\n';
                        SetCode(EPARSE);
                        IssueMessage(0,PARTIAL_PARSE_FAILURE,spritenum--);
                    }
                }
            }
        }
        if(peek(in)==EOF){
            flush_buffer();
            (*real_out)<<NFO_HEADER(NFOversion);
            if(isPatch)(*real_out)<<"    0 * 4\t "<<mysprintf("%8x\n",GetState(DIFF)?0:spritenum);
            (*real_out)<<outbuffer.str();
            pNfo=real_out;
            final_sanity();
            return 0;
        }
        inj_getline(in,sprite);
    }
}

bool verify_real(string&data){
    string::size_type loc=NPOS;
    string udata=UCase(data);
    while(true){
        loc=udata.find(".PCX",loc+1);
        if(loc==NPOS){
            IssueMessage(0,REAL_NO_FILENAME);
            return false;
        }
        if(isspace(data[loc+4]))break;
    }
    string curfile=data.substr(0,loc+4);
    static string prevfile;
    static int prevy=0;
    if(curfile!=prevfile){
        prevy=0;
        prevfile=curfile;
    }
    int var_list[9]={0,0,1,1,1,0,0};
    int&xpos=var_list[0],&ypos=var_list[1],&comp=var_list[2],&ysize=var_list[3],&xsize=var_list[4],&xrel=var_list[5],
        &yrel=var_list[6];
    const char*const format_list[7]={
        "%d%s",
        "%d %d%s",NULL,
        "%d %d %2x %d%s",
        "%d %d %2x %d %d%s",
        "%d %d %2x %d %d %d%s",
        "%d %d %2x %d %d %d %d%s"};
    const char*const var_names[7]={"xpos","ypos","comp","ysize","xsize","xrel","yrel"};
    string meta=data.substr(loc+5);
    size_t offs=NPOS;
    uchar state=0; // bitmask of calculated variables
    int var;
    while((var=sscanf(meta.c_str(),"%d %d %2x %d %d %d %d",&xpos,&ypos,&comp,&ysize,&xsize,&xrel,&yrel))!=7){
        if(RPNOFF()||
            //Calculation on comp, already tried to calculate this datum, or no more parens->broken sprite
            var==2||state&(1<<var)||(offs=meta.find('('))==NPOS||
            //If OK so far, calculate and store datum. If calc fails, offs == NPOS; sprite is broken
            (var_list[var]=DoCalc(meta,offs),offs==NPOS)){
            if(COMMENTON()||RPNON()||(RPNOFF()&&meta.find('(')==NPOS))IssueMessage(0,REAL_MISSING_DATA,var_names[var]);
            return COMMENTOFF();
        }
        state|=1<<var;//Mark calculation
        // **FIXME** 64-bit safety (?) (next 2 lines)
        //store pointer to remainder of metadata immediately after calculated datum, and ...
        (const char*&)(var_list[var+1])=meta.c_str()+offs;
        //... pass it as one or two 32-bit integers, to be read with va_arg(ap,char*).
        meta=mysprintf(format_list[var],xpos,ypos,comp,ysize,xsize,xrel,yrel,var_list[7],var_list[8]);
    }
    if(state)data=data.substr(0,loc+5)+meta;
    if(xpos<0)IssueMessage(ERROR,REAL_VAL_TOO_SMALL,"xpos",0);
    if(ypos<0)IssueMessage(ERROR,REAL_VAL_TOO_SMALL,"ypos",0);
    else if(ypos<prevy)IssueMessage(WARNING1,REAL_MOVES_UP);
    if(!(comp&1)||(comp&0x0B)!=comp)IssueMessage(comp==0xFF?ERROR:WARNING1,REAL_BAD_COMP,comp);
    if(xsize<1)IssueMessage(ERROR,REAL_VAL_TOO_SMALL,"xsize",1);
    else if(xsize>0xFFFF)IssueMessage(ERROR,REAL_VAL_TOO_LARGE,"xsize",0xFFFF);
    if(ysize<1)IssueMessage(ERROR,REAL_VAL_TOO_SMALL,"ysize",1);
    else if(ysize>0xFF)IssueMessage(ERROR,REAL_VAL_TOO_LARGE,"ysize",0xFF);
    if(xsize*ysize>0xFFFF)IssueMessage(ERROR,REAL_SPRITE_TOO_LARGE);
    if(xrel<-32768)IssueMessage(ERROR,REAL_VAL_TOO_SMALL,xrel,-32768);
    else if(xrel>32767)IssueMessage(ERROR,REAL_VAL_TOO_LARGE,xrel,32767);
    if(yrel<-32768)IssueMessage(ERROR,REAL_VAL_TOO_SMALL,yrel,-32768);
    else if(yrel>32767)IssueMessage(ERROR,REAL_VAL_TOO_LARGE,yrel,32767);
    return true;
}

string smash(const string&sprite,int&err){
    istringstream in(sprite);
    ostringstream out;
    char ch;
    bool literal=false;
    err=0;
    while(in.get(ch)){

//This section is a rewrite of Petr Baudis's (Pasky's) patch to info.cc:
//http://pasky.or.cz/~pasky/dev/openttd/grfcodec/grfcodec+unix2.patch
// now incorporated into GRFCodec 0.9.6 and later
//That patch is Copyright Petr Baudis
        if(ch=='"'){
            if(out.str().length()%2){
                err=OFFSET_STR;
                return "";
            }
            SetVersion(5);
            literal=!literal;
            continue;
        }
        if(literal){
            /*TODO: Uncomment if/when GRFCodec DTRT.
            if(ch=='\\'){
                in.get(ch);
                if(ch=='\n')continue;
                if(ch=='x'||ch=='X'){
                    in.get(ch);
                    out.put(ch);
                    in.get(ch);
                    out.put(ch);
                    continue;
                }//*/
            out<<itoa((unsigned char)ch,16,2);
            continue;
        }
 
        
        if(COMMENT.find(ch)!=NPOS){
            in.ignore(INT_MAX,'\n');
            continue;
        }
        if(isspace(ch))continue;
        if(string(VALID_PSEUDO).find(ch)==NPOS){
            err=INVALID_CHAR;
            return string(1,ch);
        }
        out<<(char)toupper(ch);
    }
    if(literal)err=UNTERM_STR;
    return out.str();
}

void output_buffer(const string&sprite,int&spritenum,bool isPatch){
    int err;
    string data=smash(sprite,err);
    switch(err){
    case 0:break;
    case INVALID_CHAR:
        IssueMessage(0,PARSE_FAILURE,spritenum);
        IssueMessage(0,INVALID_CHARACTER,data[0]);
        comment_sprite(sprite);
        SetCode(EPARSE);
        return;
    case UNTERM_STR:
        IssueMessage(0,PARSE_FAILURE,spritenum);
        IssueMessage(0,UNTERMINATED_STRING);
        comment_sprite(sprite);
        SetCode(EPARSE);
        return;
    case OFFSET_STR:
        IssueMessage(0,PARSE_FAILURE,spritenum);
        IssueMessage(0,OFFSET_STRING);
        comment_sprite(sprite);
        SetCode(EPARSE);
        return;
    DEFAULT(output_buffer,err)
    }
    if(data.length()%2){
        IssueMessage(0,PARSE_FAILURE,spritenum);
        IssueMessage(0,BAD_SIZE);
        comment_sprite(sprite);
        SetCode(EPARSE);
        return;
    }
    if(data.length()){
        spritenum++;
        if(isPatch)check_sprite(data,spritenum);
        (*pNfo)<<setw(5)<<(GetState(DIFF)?-1:spritenum)<<" * "<<(GetState(DIFF)?0:(uint)data.length()/2)<<"\t ";
    }
    (*pNfo)<<sprite;
}

void comment_sprite(const string&data){
    istringstream datastream(data);
    string line;
    getline(datastream,line);
    (*pNfo)<<COMMENT_PREFIX<<"    0 * 0\t "<<line<<endl;
    while(getline(datastream,line)){
        if(!is_comment(line))(*pNfo)<<COMMENT_PREFIX;
        (*pNfo)<<line<<endl;
    }
}
